from threading import Thread
import socket
import time
import random
import sys
import os
import os.path
import subprocess
import shlex
import re

def lireligne(socket):
    """ fonction pompée sur internet pour lire ligne par ligne en
    utilisant les générateurs [[https://stackoverflow.com/questions/822001/python-sockets-buffering]] """

    buffering = True
    buffer = ""
    try:
        buffer = socket.recv(4096).decode('utf-8')
    except ConnectionResetError:
        print("socket réinitialisée")
        buffering=False
        yield ""
        # print ("verif='{}'".format(buffer))
    while buffering:
        if "\n" in buffer:
            (line, buffer) = buffer.split("\n", 1)
            yield line + "\n"
        else:
            try:
                more = socket.recv(4096).decode('utf-8')
            except ConnectionResetError:
                print("socket réinitialisée")
                more = None
            if not more:
                buffering = False
            else:
                buffer += more
    if buffer:
        yield buffer

        

class ServeurTest(Thread):
    """ Une petite classe de serveur qui lit des ligne et écrit en réponse
    le texte proposé dans son shema"""
    
    def __init__(self, host, port, shema):
        Thread.__init__(self)
        self.host = host
        self.port = port
        self.shema = shema
        self.as_client = False
        self.client = ""
        self.portclient = ""
        
        self.ssock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        self.ssock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.ssock.bind((self.host,self.port))
        self.ssock.listen(42)
        
    def run(self):
        (self.sock, (self.client, self.port, flowinfo, scopid)) = self.ssock.accept()
        print ("Un client {} {}".format(self.client, self.port))
        self.as_client = True
        self.sock.setblocking(True)

        gen = lireligne(self.sock)

        shem = self.shema[0]
        self.envoie_instable(shem)
        for shem in self.shema[1:]:
            self.lignerecue = next(gen)
            print ("Recu = '{}'".format(self.lignerecue))
            self.envoie_instable(shem)

        self.sock.close()

            
    def envoie_instable(self, msg):
        if not self.as_client:
            raise ValueError("Le client n'est pas connecté")

        lignes = msg.split("\n")
        for lig in lignes:
            self.sock.send((lig+"\n").encode('utf-8'))
            time.sleep(0.1*random.random())

def RecupReponse(rep):
    """ Récupère les lignes du résultat commançant par ##"""
    lignes = rep.split(b"\n")
    res = [];
    for li in lignes:
        strli = li.decode("utf-8")
        if re.match("^##.*", strli):
           res.append(strli)
    return res


#class ClientTest(Thread):
#    def __init__(self, commande):
#        Thread.__init__(self)
#        self.commande = commande

def Test1(nom):
    print("\033[34;11m"
        + "#########################################\n"
        + "  Test 1 : vérification des arguments    \n"
        + "#########################################\n"
        + "\033[00m"
        );

    print("Test de la commande ./{}".format(nom))
    commande = "./{}".format(nom);
    proc = subprocess.Popen(commande, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = proc.communicate()
    good = False
    for o in out:
        if o is None:
            break
        lignes = RecupReponse(o)
        for li in lignes:
            if re.match("^## usage.*{}.*".format(nom), li):
                print ("Ligne '{}' trouvée".format(li))
                good = True
                break
        if (not good):
            print ("Le programme lancer avec de mauvais arguments devrait écrire ## usage ...")
            return False
    
    print("Test de la commande ./{} toto".format(nom))
    commande = "./{} toto".format(nom);
    args = shlex.split(commande)
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = proc.communicate()
    good = False
    for o in out:
        if o is None:
            break
        # print("o='{}'".format(o))
        lignes = RecupReponse(o)
        for li in lignes:
            if re.match("^## usage.*{}.*".format(nom), li):
                print ("Ligne '{}' trouvée".format(li))
                good = True
                break
        if (not good):
            print ("Le programme lancé avec de mauvais arguments devrait écrire ## usage ...")
            return False
    print("Test de la commande ./{} toto tata titi tutu plop".format(nom))
    commande = "./{} toto tata titi tutu plop".format(nom);
    args = shlex.split(commande)
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out = proc.communicate()
    good = False
    for o in out:
        if o is None:
            break
        # print("o='{}'".format(o))
        lignes = RecupReponse(o)
        for li in lignes:
            if re.match("^## usage.*{}.*".format(nom), li):
                print ("Ligne '{}' trouvée".format(li))
                good = True
                break
        if (not good):
            print ("Le programme lancé avec de mauvais arguments devrait écrire ## usage ...")
            return False
        
    return True


def Test2(nom):
    print("\033[34;11m"
        + "#########################################\n"
        + "  Test 2 : connexion à un serveur        \n"
        + "#########################################\n"
        + "\033[00m"
        );
    port = random.randrange(2000, 3000, 1)
    serv = ServeurTest("", port, ["202 Coucou"])
    serv.start()
    args = ["./{}".format(nom), "toto", "tata", "::1", "{}".format(port)]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
#    out = proc.communicate()
#    for o in out:
#        print("o='{}'".format(o))
    time.sleep(1)
    if serv.as_client:
        print ("La connexion au serveur local sur le port {} est possible\n".format(port))
        return True
    else:
        print ("La connexion au serveur local sur le port {} est impossible\n".format(port))
        print ("Vérifier la bonne lecture du port et l'utilisation d'IPV6")        
        return False

def Test3(nom):
    print("\033[34;11m"
        + "#########################################\n"
        + " Test 3 : lecture de réponse du serveur  \n"
        + "#########################################\n"
        + "\033[00m"
        );
    port = random.randrange(2000, 3000, 1)
    repserv = "202-Coucou\nune série de ligne\npour rendre la lecture difficile\n333_afin de voir \n22  si vos fonctions de lectures\n1   sont efficaces.\n200 bravo\n"
    serv = ServeurTest("", port, [repserv])
    serv.start()
    args = ["./{}".format(nom), "toto", "tata", "::1", "{}".format(port)]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    out = proc.communicate()
    lignes = RecupReponse(out[0]);
    rep1 = "";
    for li in lignes:
        print("lu='{}'".format(li))
        m = re.match("^##1-: (.*)", li)
        if m:
            rep1 = rep1 + m.group(1)+"\n"; 
    print ("rep1 = {}".format(rep1))
    if (rep1 == repserv):
        print ("La première réponse est bien lue")
        return True
    else:
        print ("L'affichage de la première réponse donne :{}".format(rep1))
        print ("Il devrait donner :{}".format(repserv))
        return False
    return True

def Test4(nom):
    print("\033[34;11m"
        + "#########################################\n"
        + " Test 4 : vérification des premier échanges\n"
        + "#########################################\n"
        + "\033[00m"
        );
    port = random.randrange(2000, 3000, 1)
    repserv = ["202 Coucou\r\n", "202-rien\r\n400 aie", ]
    serv = ServeurTest("", port, repserv)
    serv.start()
    args = ["./{}".format(nom), "toto", "tata", "::1", "{}".format(port)]
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    out = proc.communicate()
    lignes = RecupReponse(out[0]);
    rep1 = "";
    serv.join()
    if not re.search("^HELO " ,serv.lignerecue):
        print("ERREUR la première ligne envoyée devrait commencée par HELO")
        print("  à la place c'est '{}'".format(serv.lignerecue))
        return False
    
    good = False
    affclient = ""
    for li in lignes:
        print("lu='{}'".format(li))
        m = re.match("^##2.?: Réponse du serveur KO", li)
        if m:
            good = True
        m = re.match("^##2.?: (.*)", li)
        if m:
            affclient = affclient + m.group(1)+"\n"
        
    if not good:
        print ("Le serveur a répondu 400 la seconde fois, le client aurait du le détecter.")
        print ("Le client à affiché : {}".format(affclient))        
        return False
    return True

if __name__ == "__main__":
#    print ("val={}".format())
    if len(sys.argv) != 2:
        sys.stderr.write("Usage : python3 {} <nom_executable>\n".format(sys.argv[0]))
        exit(1)
    if (not os.path.isfile(sys.argv[1])):
        sys.stderr.write("Le fichier {} n'existe pas\n".format(sys.argv[1]))
        exit(1)

    nom = sys.argv[1]
    if (Test1(nom)):
        print ("\033[32;11m############# GOOD ###########\033[00m")
    else:
        print ("\033[31;11mFAIL\033[00m")
        exit(1)

    if (Test2(nom)):
        print ("\033[32;11m############# GOOD ###########\033[00m")
    else:
        print ("\033[31;11mFAIL\033[00m")
        exit(1)

    if (Test3(nom)):
        print ("\033[32;11m############# GOOD ###########\033[00m")
    else:
        print ("\033[31;11mFAIL\033[00m")
        exit(1)

    if (Test4(nom)):
        print ("\033[32;11m############# GOOD ###########\033[00m")
    else:
        print ("\033[31;11mFAIL\033[00m")
        exit(1)

    print("Tous les tests programmés sont passé, vous pouvez essayer sur le serveur jouet.")
    # serv = ServeurTest("", 8080, ["202-Coucou\nune série de ligne\npour rendre la lecture difficile\n333_afin de voir \n22  si vos fonctions de lectures\n1   sont efficaces.\n200 bravo\n"])
    # serv.start()
    # serv.join()

